/* knetUdc -- install udc i/o functions in knetfile interface in Heng Li's samtools lib. */
/* As of 2/23/10, the KNETFILE_HOOKS extension is a UCSC-local modification of samtools. */

#ifndef KNETUDC_H
#define KNETUDC_H

void knetUdcInstall();
/* install udc i/o functions in knetfile interface in Heng Li's samtools lib. */

#endif//ndef KNETUDC_H
